from pyocd.flash.file_programmer import FileProgrammer
from pathlib import Path
import numpy as np
import struct
from typing import Sequence
from typing import Union
from typing import Callable

from riotee_probe.protocol import *
from riotee_probe.session import RioteeProbeSession
from riotee_probe.intelhex import IntelHex16bitReader


class Target(object):
    def __init__(self, session: RioteeProbeSession):
        self._session = session

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass

    def halt(self):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def resume(self):
        raise NotImplementedError

    def write(self, addr, data):
        raise NotImplementedError

    def read(self, addr, n: int):
        raise NotImplementedError

    def program(self, fw_path: Path, progress: Callable = None):
        raise NotImplementedError


class TargetMSP430(Target):
    def __enter__(self):
        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_CONNECT)
        return self

    def __exit__(self, *exc):
        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_DISCONNECT)

    def reset(self):
        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_RESET)

    def resume(self):
        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_RESUME)

    def halt(self):
        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_HALT)

    def write(self, addr: int, data: Union[Sequence[np.uint16], np.uint16]):
        if hasattr(data, "__len__"):
            pkt = struct.pack(f"=IB{len(data)}H", addr, len(data), *data)
        else:
            pkt = struct.pack(f"=IBH", addr, 1, data)

        # One Byte is required for request type
        if len(pkt) >= DAP_VENDOR_MAX_PKT_SIZE:
            raise ValueError("Data length exceeds maximum packet size")

        rsp = self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_WRITE, pkt)

    def read(self, addr, n_words: int = 1):
        pkt = struct.pack(f"=IB", addr, n_words)
        rsp = self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_SBW_READ, pkt)
        rsp_arr = np.frombuffer(rsp, dtype=np.uint16)
        if n_words == 1:
            return rsp_arr[0]
        else:
            return rsp_arr

    def program(self, fw_path: Path, progress: Callable = None, verify: bool = True):
        ih = IntelHex16bitReader()
        ih.loadhex(fw_path)

        self.halt()
        # Overhead: 1B request, 4B address, 1B len -> 6B
        pkts = list(ih.iter_packets((DAP_VENDOR_MAX_PKT_SIZE - 6) // 2))

        for i, pkt in enumerate(pkts):
            self.write(pkt.address, pkt.values)
            if verify:
                rb = self.read(pkt.address, len(pkt))
                if (rb != pkt.values).any():
                    raise Exception(f"Verification failed at 0x{pkt.address:08X}!")
            if progress:
                progress((i + 1) / len(pkts))

        self.resume()


class TargetNRF52(Target):
    def __enter__(self):
        self._session._board.init()
        self._session._inited = True
        return self

    def __exit__(self, *exc):
        try:
            self._session._board.uninit()
        except Exception as e:
            print("Error during board uninit:", e)
        self._session._inited = False

    def program(self, fw_path: Path, progress: Callable = None):
        if progress is None:

            def progress(arg):
                pass

        FileProgrammer(self._session, progress=progress).program(str(fw_path))

    def halt(self):
        self._session.board.target.halt()

    def reset(self):
        self._session.board.target.reset()

    def resume(self):
        self._session.board.target.resume()

    def write(self, addr, data: Union[Sequence[np.uint32], np.uint32]):
        if hasattr(data, "__len__"):
            self._session.board.target.write_memory_block32(addr, data)
        else:
            self._session.board.target.write_memory(addr, data)

    def read(self, addr, n_words: int = 1):
        if n_words == 1:
            return self._session.board.target.read_memory(addr)
        else:
            return self._session.board.target.read_memory_block32(addr, n_words)



"""Riotee Probe Python package"""

__version__ = "1.0.2"

from riotee_probe.probe import RioteeProbe
from riotee_probe.probe import RioteeProbeProbe
from riotee_probe.probe import RioteeProbeBoard
from riotee_probe.target import TargetNRF52
from riotee_probe.target import TargetMSP430
from riotee_probe.probe import GpioDir
from riotee_probe.probe import get_connected_probe



from contextlib import contextmanager
import click
from progress.bar import Bar
from riotee_probe.probe import get_connected_probe
from riotee_probe.probe import GpioDir

device_option = click.option("-d", "--device", type=click.Choice(["msp430", "nrf52"]), default="nrf52")


@contextmanager
def get_target(device):
    with get_connected_probe() as probe:
        if device == "nrf52":
            target_gen = probe.nrf52
        else:
            target_gen = probe.msp430

        with target_gen() as target:
            yield target


@click.group
def cli():
    pass


@cli.command(short_help="Control power supply bypass (Board only)")
@click.option("--on/--off", required=True)
@click.pass_context
def bypass(ctx, on):
    with get_connected_probe() as probe:
        probe.bypass(on)


@cli.command(short_help="Control target power supply")
@click.option("--on/--off", required=True)
@click.pass_context
def target_power(ctx, on):
    with get_connected_probe() as probe:
        probe.target_power(on)


@cli.command
@device_option
@click.option("-f", "--firmware", type=click.Path(exists=True), required=True)
def program(device, firmware):
    with get_target(device) as target:
        bar = Bar("Uploading..", max=100)

        def update_bar(fraction):
            bar.goto(100 * fraction)

        target.program(firmware, progress=update_bar)
        bar.finish()


@cli.command
@device_option
def reset(device):
    with get_target(device) as target:
        target.reset()


@cli.command
@device_option
def halt(device):
    with get_target(device) as target:
        target.halt()


@cli.command
@device_option
def resume(device):
    with get_target(device) as target:
        target.resume()


@cli.group(short_help="Control GPIOs (Probe only)")
def gpio():
    pass


@gpio.command(short_help="Configure GPIO pin direction")
@click.option("--pin-no", "-p", type=int, required=True, help="Pin number")
@click.option(
    "--direction",
    "-d",
    type=click.Choice(["in", "out"], case_sensitive=False),
    required=True,
    help="Pin direction",
)
def dir(pin_no, direction):
    with get_connected_probe() as probe:
        if direction == "in":
            probe.gpio_dir(pin_no, GpioDir.GPIO_DIR_IN)
        else:
            probe.gpio_dir(pin_no, GpioDir.GPIO_DIR_OUT)


@gpio.command(short_help="Set GPIO pin")
@click.option("--pin-no", "-p", type=int, required=True, help="Pin number")
@click.option(
    "--state",
    "-s",
    type=click.Choice(["high", "1", "low", "0"], case_sensitive=False),
    required=True,
    help="Pin state",
)
def set(pin_no, state):
    with get_connected_probe() as probe:
        if state in ["high", "1"]:
            probe.gpio_set(pin_no, True)
        else:
            probe.gpio_set(pin_no, False)


@gpio.command(short_help="Read GPIO pin")
@click.option("--pin-no", "-p", type=int, required=True, help="Pin number")
def get(pin_no):
    with get_connected_probe() as probe:
        state = probe.gpio_get(pin_no)
        click.echo(state)


if __name__ == "__main__":
    cli()



from enum import IntEnum


class DapRetCode(IntEnum):
    DAP_OK = 0
    DAP_ERROR = 0xFF


ID_DAP_VENDOR0 = 0x80

DAP_VENDOR_MAX_PKT_SIZE = 64


class ReqType(IntEnum):
    ID_DAP_VENDOR_VERSION = 0x80
    ID_DAP_VENDOR_POWER = 0x81
    ID_DAP_VENDOR_SBW_CONNECT = 0x82
    ID_DAP_VENDOR_SBW_DISCONNECT = 0x83
    ID_DAP_VENDOR_SBW_RESET = 0x84
    ID_DAP_VENDOR_SBW_HALT = 0x85
    ID_DAP_VENDOR_SBW_RESUME = 0x86
    ID_DAP_VENDOR_SBW_READ = 0x87
    ID_DAP_VENDOR_SBW_WRITE = 0x88
    ID_DAP_VENDOR_GPIO_SET = 0x89
    ID_DAP_VENDOR_GPIO_GET = 0x8A
    ID_DAP_VENDOR_BYPASS = 0x8B


class BypassState(IntEnum):
    BYPASS_OFF = 0
    BYPASS_ON = 1


class TargetPowerState(IntEnum):
    TARGET_POWER_OFF = 0
    TARGET_POWER_ON = 1


class IOSetState(IntEnum):
    IOSET_OUT_LOW = 0
    IOSET_OUT_HIGH = 1
    IOSET_IN = 2


class IOGetState(IntEnum):
    IOGET_LOW = 0
    IOGET_HIGH = 1



from enum import Enum
import struct
from contextlib import contextmanager

from riotee_probe.protocol import ReqType
from riotee_probe.protocol import IOSetState
from riotee_probe.session import RioteeProbeSession
from riotee_probe.target import TargetNRF52
from riotee_probe.target import TargetMSP430


class GpioDir(Enum):
    GPIO_DIR_IN = 0
    GPIO_DIR_OUT = 1


@contextmanager
def get_connected_probe():
    with RioteeProbeSession() as session:
        if session.product_name == "Riotee Board":
            yield RioteeProbeBoard(session)
        elif session.product_name == "Riotee Probe":
            yield RioteeProbeProbe(session)
        else:
            raise Exception(f"Unsupported probe {session.product_name} selected")


class RioteeProbe(object):
    def __init__(self, session):
        self._session = session

    @contextmanager
    def msp430(self):
        with TargetMSP430(self._session) as msp430:
            yield msp430

    @contextmanager
    def nrf52(self):
        with TargetNRF52(self._session) as nrf52:
            yield nrf52

    def target_power(self, state: bool):
        pkt = struct.pack("=B", state)
        return self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_POWER, pkt)

    def gpio_set(self, pin: int, state: bool):
        raise NotImplementedError

    def gpio_get(self, pin: int):
        raise NotImplementedError

    def bypass(self, state):
        raise NotImplementedError

    def fw_version(self) -> str:
        ret = self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_VERSION)
        return str(ret, encoding="utf-8")


class RioteeProbeProbe(RioteeProbe):
    def gpio_set(self, pin: int, state: bool):
        if state:
            pkt = struct.pack("=BB", pin, IOSetState.IOSET_OUT_HIGH)
        else:
            pkt = struct.pack("=BB", pin, IOSetState.IOSET_OUT_LOW)

        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_GPIO_SET, pkt)

    def gpio_get(self, pin: int) -> bool:
        pkt = struct.pack("=B", pin)

        rsp = self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_GPIO_GET, pkt)
        return bool(rsp[0])

    def gpio_dir(self, pin: int, dir: GpioDir):
        if dir == GpioDir.GPIO_DIR_IN:
            pkt = struct.pack("=BB", pin, IOSetState.IOSET_IN)
        else:
            pkt = struct.pack("=BB", pin, IOSetState.IOSET_OUT_LOW)

        self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_GPIO_SET, pkt)


class RioteeProbeBoard(RioteeProbe):
    def bypass(self, state: bool):
        pkt = struct.pack("=B", state)
        return self._session.vendor_cmd(ReqType.ID_DAP_VENDOR_BYPASS, pkt)



from intelhex import IntelHex
import numpy as np


class PacketOverflowException(Exception):
    pass


class Packet16bit(object):
    def __init__(self, address: int, max_size: int):
        self._max_size = max_size
        self._buffer = list()
        self.address = address

    def append(self, value: int):
        if len(self._buffer) < self._max_size:
            self._buffer.append(value)
        else:
            raise PacketOverflowException("Packet buffer full")

    def __len__(self):
        return len(self._buffer)

    @property
    def values(self):
        return np.array(self._buffer, dtype=np.uint16)


class IntelHex16bitReader(IntelHex):
    def __init__(self, source=None):
        super().__init__(source)

    def iter_words(self):
        """Iterates segments and yields addresses and corresponding 16-bit values."""
        for addr, addr_stop in self.segments():
            while addr < addr_stop:
                value = self[addr] + (self[addr + 1] << 8)
                yield addr, value
                addr += 2

    def iter_packets(self, pkt_max_size):
        """Iterates segments and yields packets of continuous data with specified maximum size."""
        for addr, addr_stop in self.segments():
            pkt = Packet16bit(addr, pkt_max_size)
            while addr < addr_stop:
                value = self[addr] + (self[addr + 1] << 8)
                try:
                    pkt.append(value)
                except PacketOverflowException:
                    yield pkt
                    pkt = Packet16bit(addr, pkt_max_size)
                    pkt.append(value)
                addr += 2
            yield pkt



from contextlib import contextmanager
from pyocd.core.helpers import ConnectHelper
from pyocd.core.session import Session
from riotee_probe.protocol import DapRetCode
from riotee_probe.protocol import ID_DAP_VENDOR0


class RioteeProbeSession(Session):
    def __init__(self):
        self.product_name = None

    def __enter__(self):
        probe = ConnectHelper.choose_probe()
        super().__init__(probe, target_override="nrf52")
        self.open(init_board=False)
        self.product_name = probe.product_name
        return self

    def __exit__(self, *exc):
        self.close()

    def vendor_cmd(self, cmd_id: int = None, data: bytes = None) -> bytes:
        # Subtract command offset here, will be added again later
        cmd_id = cmd_id - ID_DAP_VENDOR0
        rsp = self.probe._link.vendor(cmd_id, data)
        if rsp[0] != DapRetCode.DAP_OK:
            raise Exception(f"Probe returned error code {rsp[0]}")
        return bytes(rsp[1:])
